/* $Workfile: T2p5x4Hdlc.c $       $Revision: 45 $      $Modtime: 3/13/02 5:55p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X4HDLC.C

    A simple example console application demonstrating the use of OTX Driver.

    This example program can be used with the following boards:

      THOR-2-PCMCIA board
      THOR-2-PCI board equipped with a Vidar-5x4-ASM Daughter Board.
      THOR-8-PCI board equipped with a Vidar-5x4-ASM Daughter Board.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures all available E1T1 devices on the board. The E1T1 spans should be
    looped with an external loop cable (e.g. Li0<->Li1, Li2<->Li3, etc).
    
    The  controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    The HDLC Sender/Receiver is hosted by the DSP device. Currently it is hardcoded
    to use only one HDLC channel per Li, but it could fairly easily be expanded
    to use more than one channel per Li (see coHdlcPerLi).

    Designed for 4 and 2 DSPs (4 for Vidar-5x5-ASM and 2 for Thor-2-PCMCIA).

    SS7 FISU filtering is supported (see the command line switches).

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call
#define DEMO_LOGIC_DEVS_PER_DSP 4

#define DEMO_SPM1_HDLC_SENDER 1        // Comment this line to remove the HDLC Sender Logical devices from this application

#define DEMO_MAX_MULTIPLE_FRAME_READ 128


#define DEMO_DEFAULT_LOG_FILE "T2p5x4Hdlc.log"
#define DEMO_DEFAULT_BOARD_NO 0


typedef struct DemoEnvT {
  char           *logFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  int             coDsp;
  int             coLi;
  int             coHdlcPerLi;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL        bSs7FisuFiltering;
  OTX_BOOL        bHdlcBitInversion;
  int             boardNo;
  OTX_BOOL        bUseDspDebugger;
  OTX_BOOL        bSuperChannel;
  int             nTimeSlots;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[16];
OTX_UINT32  g_coHdlcBadRxFrames[16];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 0;
OtxHdlcSS7FisuS g_tFisu;
OtxHdlcSS7ShortLssuS g_tLssu;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-8-PCI)
  OTX_HANDLE hVidarAsmController0;  // Only used for Thor-2-PCI and Thor-8-PCI (not for Thor-2-PCMCIA)
  OTX_HANDLE hLi[8];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_INT32  nLi;
  OTX_INT32  nDsp;
} g_sqDspHdlcSender[32];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_INT32  nLi;
  OTX_INT32  nDsp;
} g_sqDspHdlcReceiver[32];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (result == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
    // On the Thor-2-PCI and the Thor-2-PCMCIA board the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < 2; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
    {
      OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI

      // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Open physical DSP devices
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    // Get handles to the physcial DSP devices #0 and #1
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_8_PCI:
    
    // Get a handle to to the ASM Daughter Board
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_VIDAR_5x_ASM, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

	// Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    DemoLogf("Loading OTX_DSP_SPM_OTXHDLC to DSP#%d \n", i);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXHDLC); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
  if (g_demoEnv.bUseDspDebugger) {
    DemoLogf("Please load and run DSP program %s to DSP#%d-%d in your DSP Debugger\n", "OtxHdlc.out", 0, g_demoEnv.coDsp-1);
    DemoLogf("Press any key when all the DSP program are running...\n");
    getch();
    
    // At this point we need to tell the host to restart the programs 
    // (to issue the initial IO commands to the DSP again)
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      DemoLogf("Restarting program on DSP#%d\n", i);
      result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
  }


}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Initialize the T1/E1 Line Interface Devices

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}

/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
    OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
#ifdef DEMO_SPM1_HDLC_SENDER
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_sqDspHdlcSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);
#endif
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[(i+2)%g_demoEnv.coDsp], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[(i+2)%g_demoEnv.coDsp], result);
  }
  
}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nLi;

  if (g_demoEnv.bSuperChannel) {
    // x*64kbps HDLC capacity (TS1-TS31 E1 only)
    OTX_INT32 j;
    OTX_INT32 noOfTimeSlots = g_demoEnv.nTimeSlots;
    i = 0;
    for (nLi = 0; nLi < g_demoEnv.coLi; nLi++, i++) {
      
      assert(g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_E1);
      
      result = OtxDrvCreatePipe(noOfTimeSlots*64, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      result = OtxDrvCreatePipe(noOfTimeSlots*64, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      
      // Timeslot on the E1 span
      memset(sqTimeSlotMask, 0x00, 32 * sizeof(OTX_UINT8));
      for (j = 0; j < noOfTimeSlots; j++) {
        sqTimeSlotMask[1+j] = 0xFF;
      }
      
      result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      
      // Timeslot on the DSP
      //memset(sqTimeSlotMask, 0xFF, 32 * sizeof(OTX_UINT8));
      //sqTimeSlotMask[0] = 0x00;
      result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[(i+2)%g_demoEnv.coDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      
      // Create the following Connections in receive direction:
      //  Li  ------> HDLC Receiver
      result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe, -1, g_sqDspHdlcReceiver[i].hHandle,       -1); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      
      // Make the following connections in transmit direction:
      // HDLC Sender   -----> Li
#ifdef DEMO_SPM1_HDLC_SENDER
      result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle, -1, g_sqDspHdlcSender[i].hPipe, -1); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
#endif
      
    }
  } else { 
    // 64kbps HDLC capacity
    OTX_INT32 j;
    OTX_INT32 ts;
    
    i = 0;
    for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
      
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        ts = 24; 
      } else {
        // E1 mode
        ts = 16; 
      }
      
      for (j = 0; j < g_demoEnv.coHdlcPerLi; j++, i++, ts--) {
        
        result = OtxDrvCreatePipe(64, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        result = OtxDrvCreatePipe(64, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
        
        // Connect pipes to the E1/T1 device
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF; // Timeslot on the E1 span
        
        g_sqDspHdlcReceiver[i].nLi = nLi;
        result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        
        g_sqDspHdlcSender[i].nLi = nLi;
        result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);

        g_sqDspHdlcReceiver[i].nDsp = (i+2)%g_demoEnv.coDsp;
        g_sqDspHdlcSender[i].nDsp = i%g_demoEnv.coDsp; 

      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
        // Connect pipes to the DSP device (Fixed timeslots since the PCMCIA board does not have a time-space switch)
        result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[g_sqDspHdlcReceiver[i].nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      } else {
        // Connect pipes to the DSP device (Arbitrary timeslot on the DSP highway)
        result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[g_sqDspHdlcReceiver[i].nDsp], -1, NULL, 0); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], -1, NULL, 0); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      }
        
        // Create the following Connections in receive direction:
        //  Li  ------> HDLC Receiver
        result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe, -1, g_sqDspHdlcReceiver[i].hHandle,       -1); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        
        // Make the following connections in transmit direction:
        // HDLC Sender   -----> Li
#ifdef DEMO_SPM1_HDLC_SENDER
        result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle, -1, g_sqDspHdlcSender[i].hPipe, -1); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
#endif
        
        DemoLogf("Hdlc Sender[%d]   connected (Dsp#%d, Li#%d) \n", i, g_sqDspHdlcSender[i].nDsp, g_sqDspHdlcSender[i].nLi);
        DemoLogf("Hdlc Receiver[%d] connected (Dsp#%d, Li#%d) \n", i, g_sqDspHdlcReceiver[i].nDsp, g_sqDspHdlcReceiver[i].nLi);
      }
    }
    
  } // g_demoEnv.bSuperChannel
      


}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {


    
#ifdef DEMO_SPM1_HDLC_SENDER
/*    {
/*
	// Specify whether the HDLC Sender device should send us the OTX_HDLC_SENDER_EC_FRAME_SENT event
      OTX_BOOL    bHdlcFrameSentEvents = OTX_TRUE;
      
      DemoLogf("HDLC Sender[%d] Setting OTX_ATTR_HDLC_SENDER_GENERATE_FRAME_SENT_EVENT=%s\n", i, (bHdlcFrameSentEvents) ? "TRUE": "FALSE");
      result = OtxDrvSetAttributeValueBool(g_sqDspHdlcSender[i].hHandle, OTX_ATTR_HDLC_SENDER_GENERATE_FRAME_SENT_EVENT, bHdlcFrameSentEvents); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
    }
*/
    // Initialize the Hdlc Sender 
    result = OtxHdlcSenderInit(
      g_sqDspHdlcSender[i].hHandle, // Handle to logical HDLC Filter Device
      /*OTX_HDLC_FFT_FLAGS*/OTX_HDLC_FFT_ALL_ONES,        // Fill character(s) to be send between HDLC frames
      1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);          
#endif
    
    // Initialize the Hdlc Receiver 
    result = OtxHdlcReceiverInit(
      g_sqDspHdlcReceiver[i].hHandle, // Handle to logical HDLC Filter Device
      /*OTX_HDLC_FFT_FLAGS*/OTX_HDLC_FFT_ALL_ONES,        // Fill character(s) to be send between HDLC frames
      1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                  // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);          
    
    if (g_demoEnv.bSs7FisuFiltering) {
      result = OtxHdlcSS7SetFilter(g_sqDspHdlcReceiver[i].hHandle, OTX_HDLC_SS7_FILTER_FISU_ALL | OTX_HDLC_SS7_FILTER_SHORT_LSSU | OTX_HDLC_SS7_FILTER_LONG_LSSU); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
    }
    
  }
}



/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
    //DemoLogf("** i=%d Enabling hSpm1Sender[%d] \n", i); getch();
#ifdef DEMO_SPM1_HDLC_SENDER
    result = OtxDrvEnable(g_sqDspHdlcSender[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
#endif
  }

  for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
    //DemoLogf("** i=%d Enabling hSpm1Receiver[%d] \n", i); getch();
#if 1
    result = OtxDrvEnable(g_sqDspHdlcReceiver[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
#endif
  }
  
}






/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
#if 0
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
#else
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_DOUBLE_FRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_DOUBLE_FRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_G775;
#endif
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000;
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OsaConsoleExit();
      OtxDrvReleaseLib();
			exit(1);
    }

  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;

  ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];

  if (g_bHdlcVerbose) {
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
  }

  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
       DemoLogf("DSP[%d] Event: %s (m_nParam=0x%X, m_nRequestId=0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);

    //	DemoLogf("DSP[%d] Event: %s (m_nParam=%d, m_nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

#if 1 // Single Frame Read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT8  buf[1024];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OtxHdlcFrameEndCodeE nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      OTX_RESULT result;
      
      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      while (otxRc == OTX_S_OK) {
        
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId]++;
        } else {
          DemoLogf("HDLC Frame Received with an invalid m_nSourceId (%d)\n", pEventData->m_nSourceId);
        }
        
        if (g_bHdlcVerbose) {
          DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          if (nFrameEndCode) {
            g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
            
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 256; i++) {
            DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
            
            if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
              DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
              result = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, result);
            }
                        
          } else {
            if ((nFrameLen != (OTX_UINT32)g_cbHdlcFrame) && (nFrameLen != 3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            } else if (memcmp(buf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            }
          }
        }
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      }
      if (otxRc != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
    } else {
      // Unexpected event from this device
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#else // Multiple frame read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
#define CB_BUF_SIZE 8192
      OTX_UINT8  buf[CB_BUF_SIZE];
      OTX_UINT8  *pBuf;
			OtxHdlcFrameInfoS *pFrame;
      OTX_UINT32 coFrames;
      OTX_UINT32 j;
      
      otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);
      while (otxRc == OTX_S_OK) {
      
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId] += coFrames;
        }

        if (g_bHdlcVerbose) {
          DemoLogf("%d HDLC Frame(s) Received, Sequence Number = %d\n", coFrames, pEventData->m_nParam);
        }
        
        pFrame = (OtxHdlcFrameInfoS *) buf;

        for (j = 0; j < coFrames; j++) {

					pBuf = pFrame->sqData;

          if (g_bHdlcVerbose) {
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->nFrameSeq, pFrame->nFrameEndCode, pFrame->nFrameSeq);
            for (i = 0; i < pFrame->cbOctets && i < CB_BUF_SIZE; i++) {
              DemoLogf("%.2X ", pBuf[i]);
            }
            DemoLogf("\n");
          } else {
            if (pFrame->nFrameEndCode) {
              
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
              }
              
              
            } else {
              if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
              DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
                
              } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
                for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                  DemoLogf("%.2X ", pBuf[i]);
                }
                DemoLogf("\n");
                
              }
            }
          }  

          // Step pointer to next frame (if any)
					pFrame = OTX_STRUCT_ADD_BYTE_OFFSET(OtxHdlcFrameInfoS, pFrame, pFrame->nOffsetNext);           }

        // Read next frame (if any)
        otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);

      } 
      if (otxRc != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadDataEx() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#endif


  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentHdlcDev = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
#ifdef DEMO_SPM1_HDLC_SENDER
    DemoLogf(" t   Send an HDLC frame from current HDLC Sender\n");
    DemoLogf(" f   Start sending FISUs (continuosly) from current HDLC Sender\n");
    DemoLogf(" i   Start sending LSSUs (continuosly) from current HDLC Sender\n");
    DemoLogf(" m   Send an MSU from current HDLC Sender\n");
    DemoLogf(" F   Toggle OTX_HDLC_SS7_FILTER_FISU_ALL filter mask on/off for current HDLC Sender\n");
#endif
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" l   Query Line Status of the two Lis\n");
    DemoLogf(" L   Set HDLC Frame length to be transmitted\n");
    DemoLogf(" +   Set number of consecutive frame to be send from the HDLC Sender\n");
    DemoLogf(" S   Set time to sleep (in ms) between transmitted HDLC Frames (%d)\n", g_nSleepTimeBetweenHdlcFrames);
    DemoLogf(" s   Display the SS7 HDLC receive status of the active HDLC Receiver\n");
    DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" y   Query HDLC Receiver Fifo diagnostics (for current HDLC device)\n");
    break;

#if 1 // Only for debugging
  case 'x':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Dump profiling data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 3);
    }
    break;
  case 'j':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Dump debug data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
    }
    break;
  case 'J':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Dump debug data for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;
  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;
  case 'k':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = 0;
      DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;
  case 'K':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = 1;
      DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;

#endif
  case 'y':
    {
      OTX_UINT32 nAttribValue;
      DemoLogf("Fifo Diagnostics for HDLC Receiver device#%d:\n", nCurrentHdlcDev);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Size (in frames)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Size (in octets)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Peak usage (in frames)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Peak usage (in octets)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Current usage (in frames): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Current usage (in octets): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_LOST_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Number of lost frames    : %d\n", nAttribValue);
    }
    break;


  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < 16; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
          g_coHdlcBadRxFrames[i] = 0;
        }
      }
    }
    break;

  case 'V':
    {
      OTX_INT32 i;
      for (i = 0; i < 16; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("Tot[%.2d]=%.6d (Bad[%.2d]=%.6d)  ", i, g_coHdlcRxFrames[i], i, g_coHdlcBadRxFrames[i]);
          if (i && i % 2) {
            DemoLogf("\n");
          }
        }
      }
      DemoLogf("\n");
    }
    break;

  case 's':
    {
      OTX_UINT32  nFisuCount;   // Number of FISUs received
      OTX_UINT32  nLssuCount;   // Number of LSSUs received
      OTX_UINT32  nMsuCount;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.
      DemoLogf("SS7 Status for HdlcReceiver #%d \n", nCurrentHdlcDev);

      result = OtxHdlcSS7GetReceiveStatus(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);

      if (OTX_SUCCESS(result)) {
        DemoLogf("FISUs received: %lu\n", nFisuCount);
        DemoLogf("LSSUs received: %lu\n", nLssuCount);
        DemoLogf("MSUs  received: %lu\n", nMsuCount);
        DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
      }

    }
    break;

  case 'L':
    {
      char strBuf[12];
      DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbHdlcFrame = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame);
      } else {
        DemoLogf("HDLC Frame Length remains unchanged\n");
      }
    }
    break;

  case 'S':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenHdlcFrames = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;

  case '+':
    {
      char strBuf[12];

      DemoLogf("Set the number of consecutive frame to sent when 't' or 'T' are pressed (%d)\n", g_coHdlcSenderFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
      } else {
        DemoLogf("Consecutive frame count remains unchanged\n");
      }
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

#if 0 // Only for debugging
  case 'd':
    DemoLogf("Dump DSP debug data\n");
    OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 0);
    break;

  case 'b':
    DemoLogf("Init Bsp\n");
    OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 1);
    break;
#endif


#ifdef DEMO_SPM1_HDLC_SENDER
  case 't':
    {
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
      clock_t startTime, stopTime;
      //static OTX_UINT16 nHdlcTaskRef = 1;

      if (g_cbHdlcFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d HDLC frame(s) from HDLC Sender #%d (len=%d) ...", g_coHdlcSenderFrames, nCurrentHdlcDev, g_cbHdlcFrame);

//      DemoLogf("Sending %d HDLC frame(s) from HDLC Sender #%d (len=%d, nTaskRef=%d)\n", g_coHdlcSenderFrames, nCurrentHdlcDev, g_cbHdlcFrame, nHdlcTaskRef);

      // Send HDLC Frame(s)
      startTime = clock();
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_sqDspHdlcSender[nCurrentHdlcDev].hHandle, g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, OTX_TIME_INFINITY); 

//        result = OtxHdlcSendData(g_sqDspHdlcSender[nCurrentHdlcDev].hHandle, g_sqHdlcFrame, g_cbHdlcFrame, nHdlcTaskRef, OTX_TIME_INFINITY); 
        if (result == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_sqDspHdlcSender[nCurrentHdlcDev].hHandle, result);
          OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
          //nHdlcTaskRef++;
        }
      }
      stopTime = clock();

      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;


  case 'T':
    {
      OTX_UINT8 *pBadMsg = (OTX_UINT8 *)"\x00\x01\x26\x3a\x4f\x02\x02\x68\x4d\x08\x03\x82\x90\x00";
      OTX_INT32 nMsgLen = 14;
      //OTX_UINT8 *pBadMsg = "\x04\x09\x7F";
      //OTX_INT32 nMsgLen = 3;
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      DemoLogf("Sending test HDLC frame from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcSendData(hHandle, pBadMsg, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'F':
    {
      OTX_HANDLE hHandle;
      OTX_UINT32 filterMask;

      hHandle = g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle;

      result = OtxHdlcSS7GetFilter(hHandle, &filterMask); DemoCheckRc(hHandle, result);
      if (filterMask & OTX_HDLC_SS7_FILTER_FISU_ALL) {
        filterMask &= ~OTX_HDLC_SS7_FILTER_FISU_ALL;
      } else {
        filterMask |= OTX_HDLC_SS7_FILTER_FISU_ALL;
      }

      DemoLogf("Setting SS7 filter mask to %.4X for HDLC Receiver #%d (Dsp#%d, Li#%d)...", filterMask, nCurrentHdlcDev, g_sqDspHdlcReceiver[nCurrentHdlcDev].nDsp, g_sqDspHdlcReceiver[nCurrentHdlcDev].nLi);

      result = OtxHdlcSS7SetFilter(hHandle, filterMask); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'f':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
      g_tFisu.fsn++;

      DemoLogf("Start Sending FISUs (fsn=%d) continously from HDLC Sender #%d (Dsp#%d, Li#%d)...", g_tFisu.fsn, nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'i':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
      g_tLssu.fsn = g_tFisu.fsn;
      g_tLssu.sf = g_tLssu.sf + 1; 
      g_tLssu.li = 1; // Short LSSU

      DemoLogf("Start Sending LSSUs (fsn=%d, sf=%d) continously from HDLC Sender #%d (Dsp#%d, Li#%d)...", g_tLssu.fsn, g_tLssu.sf, nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_tLssu, OTX_HDLC_SS7_CB_SHORT_LSSU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'm':
    {
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
      clock_t startTime, stopTime;
      OtxHdlcSS7FisuS msu;

      if (g_cbHdlcFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      if (g_cbHdlcFrame <= OTX_HDLC_SS7_CB_FISU) {
        DemoLogf("Error: Frame Length too short\n");
        break;
      }

      // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
      g_tFisu.fsn++;
      memcpy(&msu, &g_tFisu, OTX_HDLC_SS7_CB_FISU);  // Build the first 3 bytes of the MSU
      msu.li = g_cbHdlcFrame - OTX_HDLC_SS7_CB_FISU;  // Set the Length Indicator
      memcpy(g_sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
      for (i = OTX_HDLC_SS7_CB_FISU; i < g_cbHdlcFrame; i++) {     // Add the payload the HDLC frame buffer
        g_sqHdlcFrame[i + OTX_HDLC_SS7_CB_FISU] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d MSU frame(s) (len=%d) from HDLC Sender #%d  (Dsp#%d, Li#%d)...", g_coHdlcSenderFrames, g_cbHdlcFrame, nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      // Send HDLC Frame(s)
      startTime = clock();
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nCurrentHdlcDev].hHandle, g_sqHdlcFrame, g_cbHdlcFrame, &g_tFisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
        if (result == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_sqDspHdlcSender[nCurrentHdlcDev].hHandle, result);
          OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
        }
      }
      stopTime = clock();

      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;

  case 'r':
    {
      OTX_UINT8 *pRepeatFrame = (OTX_UINT8 *)"\x10\x11\x12\x13\x14";
      OTX_INT32 nMsgLen = 5;
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      DemoLogf("Start Repeat Data Frames from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStartRepeatData(hHandle, pRepeatFrame, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
  case 'R':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      DemoLogf("Stop Repeat Data Frames from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStopRepeatData(hHandle); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
#endif


  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      printf("Set Clock Source:\n");
      printf("0 = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");
      printf("1 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to FA0 receive span\n");
      printf("2 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to FA1 receive span\n");

      switch (getch() - '0') {
      case 0:
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        break;
      case 1:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
        break;
      case 2:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
        break;
      default:
        printf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        break;
      }

      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_H100, eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      // Setting the clock source also performs a reset of the E1T1 devices. Therefore we need to reconfigure the E1T1 devices here.
      DemoConfigureE1T1Devices();

      printf("Done.\n");
    }
    break;
  
  case 'n':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev++;
      if (nCurrentHdlcDev > (g_demoEnv.coLi*g_demoEnv.coHdlcPerLi)-1) {
        // Wrap 
        nCurrentHdlcDev = 0;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev--;
      if (nCurrentHdlcDev < 0) {
        // Wrap
        nCurrentHdlcDev = (g_demoEnv.coLi*g_demoEnv.coHdlcPerLi)-1;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
    }
    break;
    
  case '#':
    {
      DemoLogf("Performing an illegal operation. Crash test\n");
      memcpy(NULL, &nCurrentHdlcDev, 200);
    }
    break;
    
        
  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_PCI;
  envP->coHdlcPerLi       = 1; // for now
  envP->coLi              = 2;
  envP->bUseDspDebugger   = OTX_FALSE;
  envP->coDsp             = 4;
  envP->bHdlcBitInversion = OTX_FALSE;
  envP->bSuperChannel     = OTX_FALSE;
  envP->nTimeSlots        = 11;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
        envP->bUseDspDebugger = OTX_TRUE;
      } else if (*(argv[i]+1) == 'I') {   // HDLC Bit Inversion
        envP->bHdlcBitInversion = OTX_TRUE;
      } else if (*(argv[i]+1) == 'P') {   // Use Super-channel (higher than 64kbps capacity)
        envP->bSuperChannel = OTX_TRUE;
      } else if (*(argv[i]+1) == 'N') {   // Number of timeslots in the superchannel
        envP->nTimeSlots = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'M') {   // Li Mode
        if (*(argv[i]+2) == 'E') {   // E1 
          envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
        } else if (*(argv[i]+2) == 'T') {   // T1
          envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
        } else {
          DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
					exit(1);
        }
      } else if (*(argv[i]+1) == 'T') {   // Board Type
        if ( *(argv[i]+2) == '2') {
          envP->nBoardType = OTX_DEVICE_THOR_PCI;
          envP->coLi = 2;
          envP->coDsp = 4;
          envP->coHdlcPerLi = 12; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
        } else if ( *(argv[i]+2) == 'M') {
          envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          envP->coLi = 2;
          envP->coDsp = 2;
          //envP->coHdlcPerLi = 6; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
          envP->coHdlcPerLi = 1; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
        } else if ( *(argv[i]+2) == '8') {
          envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          envP->coLi = 8;
          envP->coDsp = 4;
          envP->coHdlcPerLi = 3; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
        
      } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
        envP->bSs7FisuFiltering = OTX_TRUE;
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

  if (g_demoEnv.bSuperChannel) {
    envP->coHdlcPerLi       = 1; // for higher capacity pipes we support only one HDLC channel per DSP
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|M|8]    : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /S           : SS7 FISU filtering active\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /D           : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /I           : Use HDLC Bit Inversion (Default is to not invert the bits)\n");
  DemoLogf(" /P           : Use HDLC Super-channels (higher than 64kbps capacity)\n");
  DemoLogf(" /N<timeslots>: Number of 64kbps timeslots in the Super-channels (only used if /P is used)\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("T2p5x4Hdlc, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("HDLC Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.bUseDspDebugger && g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    printf("Dsp Debugging option is not supported for the Thor-2-PCMCIA card\n");
    OsaConsoleExit();
		exit(-1);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  if (g_demoEnv.bSuperChannel) {
    DemoLogf("1 superchannel per DSP (%d x 64 kbps capacity)\n", g_demoEnv.nTimeSlots);
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source 
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  } else {
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  }

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      char *pszCpldBuildNumbers;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
      free(pszCpldBuildNumbers);
    }
  }
  

  DemoLogf("\nPress 'h' for help\n");

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

#if 1
  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }
#else // Just for testing. Exiting automatically after 300ms
  OsaThreadSleep(300);
#endif

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

